
import os
from pathlib import Path
import logging
from lxf.services.measure_time import measure_time_async
import lxf.settings as settings
settings.set_logging_level(logging.ERROR)
settings.enable_tqdm=False
###################################################################

logger = logging.getLogger('converter')
fh = logging.FileHandler('./logs/converter.log')
fh.setLevel(settings.get_logging_level())
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.setLevel(settings.get_logging_level())
logger.addHandler(fh)
#################################################################

from docling.document_converter import DocumentConverter
from docling_core.transforms.chunker.base import BaseChunk
from docling_core.transforms.chunker.hierarchical_chunker import HierarchicalChunker, DocChunk
from docling.datamodel.accelerator_options import AcceleratorDevice, AcceleratorOptions
from docling.datamodel.base_models import InputFormat
from docling.datamodel.pipeline_options import (
    PdfPipelineOptions,
)
from docling.document_converter import DocumentConverter, PdfFormatOption
from docling.backend.pypdfium2_backend import PyPdfiumDocumentBackend
    
async def converter_to_mardown(fullfilename:str)->tuple[int,str] :
    """
    Converti un fichier en Markdown
    return un tuple : error, Mardown | error_message si error!=0 
    """
    source = Path(fullfilename)
    if source.exists() :
        try:
            # Docling Parse with EasyOCR (default)
            # -------------------------------
            # Enables OCR and table structure with EasyOCR, using automatic device
            # selection via AcceleratorOptions. Adjust languages as needed.
            pipeline_options = PdfPipelineOptions()
            pipeline_options.do_ocr = True
            pipeline_options.do_table_structure = True
            pipeline_options.table_structure_options.do_cell_matching = True
            pipeline_options.ocr_options.lang = ["fr"]
            pipeline_options.accelerator_options = AcceleratorOptions(
                num_threads=4, device=AcceleratorDevice.AUTO
            )           
            doc_converter = DocumentConverter(
                format_options={
                    InputFormat.PDF: PdfFormatOption(pipeline_options=pipeline_options,backend=PyPdfiumDocumentBackend)
                    
                }
            )             
            doc = doc_converter.convert(source).document    
            # chunker:HierarchicalChunker = HierarchicalChunker()        
            # chunks = chunker.chunk(dl_doc=doc)
            # for i , chk in enumerate(chunks) :
            #     doc_chunk = DocChunk.model_validate(chk)
            #     print(f"Titre : {' '.join(doc_chunk.meta.headings)}")
            #     print(doc_chunk.text)
            return 0,doc.export_to_markdown()
        except Exception as e :
            logging.error(e)
            return -2, e
    else :
        error = f"Le fichier {fullfilename} n'existe pas."
        logger.error(error)
        return -1 , error
    


from typing import List
from pydantic import BaseModel,Field

from lxf.domain.loan import Emprunteur, Montant, ObjetFinancement, Preteur

class ConditionsFinancieres(BaseModel):
    """
    Docstring for ConditionsFinancieres
    """
    is_taux_variable:bool=Field(default=False)
    calcul_taux_variable:str=Field(default="")
    commission_engagement:float=Field(default=0.0)
    periodicite_commission:str=Field(default="")
    frais_dossier:Montant=Field(default=Montant())
    methode_calcul_commission_engagement:str=Field(default="")
    depassement_autorisation:str=Field(default="")
    source:list[str]=[]

class Garantie(BaseModel):
    """
    Docstring for Garantie
    """
    label:str=Field(default="")
    hauteur:Montant=Field(default=Montant())
    rang:str=Field(default="")
    cautionnaires:list[str]=[]
    engagements:list[str]=[]

class ConditionCreditAccompagnement(BaseModel):
    """
    Docstring for ConditionCreditAccompagnement
    """
    conditions:list[str]=[]
    source:list[str]=[]
    
class ConditionFinanciereAchevement (BaseModel):
    """
    Docstring for ConditionFinanciereAchevement
    """
    conditions:list[str]=[]
    source:list[str]=[]
    
class GarantieFinanciereAchevement(BaseModel) :
    """
    Docstring for GarantieFinanciereAchevement
    """
    label:str=Field(default="")
    montant:Montant = Field(default=Montant())
    taux_annuel_commission:float=Field(default=0.0)
    modalite_reglement:str=Field(default="")
    frais_dossier:Montant=Field(default=Montant())
    conditions:str = Field(default="")
    source:list[str]=[]

class ModalitesFonctionnement(BaseModel):
    """
    Docstring for ModalitesFonctionnent
    """
    source:list[str]=[]
    date_limite_validite:str=Field(default="")
    modalites:list[str]=[]
    
class LoanProposal(BaseModel) :
    """
    Docstring for LoanProposal
    """
    num_suivi:str=Field("")
    header_offre:str=Field("")
    header_ref:str=Field("")
    header_emprunteur:str=Field("")
    header_responsable_suivi:str=Field("")
    header_num_etude:str=Field("")
    date_emission_offre:str=Field("")
    preteur: Preteur = Field(default=Preteur())
    emprunteurs:List[Emprunteur]=[]  
    objets_financement:List[ObjetFinancement]=[]
    gfa:GarantieFinanciereAchevement=Field(default=GarantieFinanciereAchevement())
    conditions_financieres:ConditionsFinancieres=Field(default=ConditionsFinancieres())
    garanties:list[Garantie] = []
    condition_credit_accompagnement:ConditionCreditAccompagnement= ConditionCreditAccompagnement()
    condition_financiere_achevement:ConditionFinanciereAchevement= ConditionFinanciereAchevement()    
    modalites_fonctionnement:ModalitesFonctionnement=Field(default=ModalitesFonctionnement())    
    source:list[str]=[]


import logging
from multiprocessing import Process
import os 
from lxf.settings import get_logging_level

###################################################################

logger = logging.getLogger('odp_proposal_extractor')
fh = logging.FileHandler('./logs/odp_proposal_extractor.log')
fh.setLevel(get_logging_level())
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.setLevel(get_logging_level())
logger.addHandler(fh)
#################################################################

from lxf.services.measure_time import measure_time_async
from lxf.extractors.finance.loans.loan_proposal_extractor import LoanProposalDataExtractor
from lxf.services.try_safe import try_safe_execute_async

@measure_time_async
async def extract_data(file_path:str,**kwargs)->str|None:
    """
    Extrait les données d'une proposition d'offre de prêt 
    """
    if file_path==None: 
        logger.error("Extract Data: file_path ne peut pas etre None")
        return None
    logger.debug(f"Demande extraction de donnees pour {file_path}")
    base_url:str=kwargs.get("base_url","http://localhost:11434")
    model_instruct:str=kwargs.get("default_model_instruct","gpt-oss:latest")
    openapi_compatibility:bool = kwargs.get("openapi_compatibility",False)
    loan_proposal:LoanProposalDataExtractor= LoanProposalDataExtractor(file_path,base_url,model_instruct, openapi_compatibility)
    result= await try_safe_execute_async(logger,loan_proposal.extract_data)         
    return result


import os
import logging


from lxf.settings import get_logging_level, load_model,enable_tqdm
class CustomFormatter(logging.Formatter): 
    def format(self, record): 
        message = record.getMessage() 
        if isinstance(message, str): 
            message = message.encode('utf-8') 
            record.msg = message 
            return super().format(record)

logger = logging.getLogger('CU Extractor')
fh = logging.FileHandler('./logs/cu_extractor.log')
fh.setLevel(get_logging_level())
formatter = logging.Formatter(str('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
fh.setFormatter(formatter)
#fh.setFormatter(CustomFormatter())
logger.setLevel(get_logging_level())
logger.addHandler(fh)
#####################################################
from lxf.services.measure_time import measure_time_async
from lxf.ai.classification.classifier import extract_text_and_table_from_file
from lxf.domain.tables import lxfTable   
from lxf.utilities.text_comparator import TextComparator
from lxf.ai.text_analysis.default_text_analysis import decouper_text_par_titres
from lxf.ai.ocr.iban_ocr import do_OcrImage_from_pdf

import re



def sanitize_text(text:str) :
    """
    """
    text= text.replace('{',"").replace("}","").replace("|","").replace("«","")
    return text


def get_numero_cu(text:str)->str:
    """
    """
    regex_CU = r"CU ?[0-9]{2,3} ?[0-9]{3} ?[0-9]{2} ?[A-Za-z]?[0-9]{4,}|CU ?[0-9]?[0-9]{5} ?[0-9]{2} ?[A-Za-z]?[0-9]{4,}"
    num_cu=""
    matches = re.findall(regex_CU,text,re.MULTILINE|re.IGNORECASE)
    if matches!=None and len(matches)>0: 
        # On a trouvé un N° de CU Valide, on prend le premier trouvé
        num_cu=matches[0] 
    return num_cu      

def check_address_with_city(text:str,nlp)->tuple[bool, str,str] :
    """
    Vérifie si le format du texte est de la forme: <adresse> à <ville>
    """
    regex_adresse_avec_ville=r"(Demeur.*à)?(?P<adresse>.*)à(?P<ville>.*)\n?"
    match = re.search(regex_adresse_avec_ville, text, re.IGNORECASE)
    if match!=None :
        _rue = match.group("adresse")
        _ville = match.group("ville")
        ville=""
        rue=""
        # Verifier qu'on a bien à faire à une adresse et /ou un ville 
        txt = f"{_rue}\n{_ville}"
        doc = nlp(txt)
        for ent in doc.ents:
            if ent.label_=="ADRESSE" :
                rue = get_rue(ent.text, nlp)
            elif ent.label_=="VILLE" :
                ville = get_ville(ent.text,nlp)
        if rue!=""  or ville!="": 
            return True,rue,ville
    return False, "",""

def get_rue(text:str, nlp) -> str :
    """
    Retourne la rue issue du texte en fonction du comparateur fourni
    """
    _text=text.lower()
    rue=""
    wellknwon_street ='rue du carrouge'
    wellknwon_street2 ="rue de lattre de tassigny"
    wellknwon_street3 ="rue maréchal de lattre de tassigny"
    wellknwon_street4 ='12 rue du carrouge'
    wellknwon_street5 ='12 rue du carouge'
    if wellknwon_street in _text or wellknwon_street2 in _text or wellknwon_street3 in _text or _text in wellknwon_street4 or _text in wellknwon_street5:
        return ""
    rue_comparator:TextComparator=TextComparator(wellknwon_street,nlp)
    rue_comparator2:TextComparator=TextComparator(wellknwon_street2,nlp)
    rue_comparator3:TextComparator=TextComparator(wellknwon_street3,nlp)
    rue_comparator4:TextComparator=TextComparator(wellknwon_street4,nlp)
    rue_comparator5:TextComparator=TextComparator(wellknwon_street4,nlp)
    limit:float=0.85
    if rue_comparator.compare_to(_text)<limit and rue_comparator2.compare_to(_text)<limit and rue_comparator3.compare_to(_text)<limit and rue_comparator4.compare_to(_text)<limit and rue_comparator5.compare_to(_text)<limit:
        rue=text
        logger.debug(f"rue= {rue}" )
    return rue
        
def get_ville(text:str,nlp) ->str:
    """
    """
    ville=""           
    if len(text)<3 : return ""
    _text= text.lower()
    wellknown_city="autun"
    wellknown_city2="plan"
    if wellknown_city in _text or wellknown_city2 in _text:
        return ""
    ville_comparator:TextComparator=TextComparator(wellknown_city)            
    if ville_comparator.compare_to(_text)<0.85 :
        ville = text                                    
        logger.debug(f"ville= {ville}")    
    return ville
def get_code_postal(text:str,nlp)->str :
    """
    """    
    cp_comparator:TextComparator=TextComparator('71400',nlp)
    cp=""
    if cp_comparator.compare_to(text)<0.9 :
        cp=text
        logger.debug(f"cp= {cp}")    
    return cp
    
@measure_time_async
async def extract_cu_basic_data(file_path) ->tuple[int,str,str,str,str]:
    """
    retourne code, erreur_message, num_cu,adresse, cadastre
    """
    text:str
    tables:list[lxfTable] 
    num_cu:str=""
    num_cu_found:bool=False
    adresse:str=""
    adresse_found=False
    cadastre:str=""
    cadastre_found:bool=False
    nlp = load_model()

    # Rechercher d'abord dans les tableaux s'il y en a 
    _,tables = await extract_text_and_table_from_file(file_name=file_path,layout=True,extract_tables=True, extract_text=False)
    if tables!=None :     
        # gestion des tableaux
        for table in tables: 
            if num_cu_found and adresse_found and cadastre_found : break
            
            for row in table.rows :
                if num_cu_found and adresse_found and cadastre_found : break
                for cell in row.cells :                     
                    done:bool=False
                    if num_cu_found and adresse_found and cadastre_found : break
                    valeur = cell.value 
                    if num_cu=="" and num_cu_found==False:
                        num_cu= get_numero_cu(valeur)
                        num_cu_found = num_cu!=""                        
                        if num_cu_found : done=True     
                    if done==False :
                        doc =nlp(valeur)
                        if adresse_found==False or cadastre_found==False :
                            # L'adresse et la ville et le CP doit être dans la même cellule
                            rue:str=""
                            cp:str=""
                            ville:str=""   
                            for ent in doc.ents : 
                                if adresse_found == False :
                                    if ent.label_ =="ADRESSE" and rue=="" :                                 
                                        # il faut gérer les adresses qui contiennent la ville :
                                        # 147 rue des Lilas à Aix-en-Provence
                                        is_city_included, _rue,_ville = check_address_with_city(ent.text, nlp)                              
                                        if is_city_included : 
                                            if _rue!="" : rue =_rue
                                            if _ville!="" : ville = _ville
                                        else :
                                            rue=get_rue(ent.text,nlp)
                                    elif ent.label_=="CODE_POSTAL" and cp=="" :                                 
                                        cp=get_code_postal(ent.text,nlp)
                                    elif ent.label_=='VILLE' and ville =="":
                                        ville = get_ville(ent.text, nlp)                            
                                if ent.label_ == "CADASTRE" and cadastre_found==False:
                                        cadastre=ent.text
                                        cadastre_found=True
                            # L'adresse et la ville et le CP doit être dans la même cellule
                            if adresse_found == False and rue!="" or ville!="" or cp!="":
                                # qu'a-t-on trouvé ? 
                                adresse = f'{rue} {cp} {ville}'.strip()
                                if adresse!="" :
                                    logger.debug(f"Adresse= {adresse}") 
                                    adresse_found=True                   
                            if cadastre_found : 
                                logger.debug(f"Cadastre= {cadastre}")
    # Effectuer une reconnaissance par traitement image 
    if num_cu_found==False or adresse_found == False  or cadastre_found==False :
        pages = await do_OcrImage_from_pdf(file_path,min_high=20,threshold_limit=120,kernel_blur_size=(15,5), kernel_dilated_size=(25,5))
        if pages!=None and len(pages)>0 :
            rue:str=""
            cp:str=""
            ville:str=""   
            for rois in pages :
                for text in rois:              
                    text = sanitize_text(text)
                    done:bool=False
                    if num_cu=="" and num_cu_found==False:
                        num_cu= get_numero_cu(text)
                        num_cu_found = num_cu!=""  
                        if num_cu_found : 
                            done=True    
                    if done==False :
                        doc =nlp(text)  
                        if adresse_found==False or cadastre_found==False :
                                    
                            for ent in doc.ents :                             
                                if adresse_found == False :
                                    if ent.label_ =="ADRESSE" and rue=="" :   
                                        # il faut gérer les adresses qui contiennent la ville :
                                        # 147 rue des Lilas à Aix-en-Provence
                                        is_city_included, _rue,_ville = check_address_with_city(ent.text, nlp)                              
                                        if is_city_included : 
                                            if _rue!="" : rue =_rue
                                            if _ville!="" : ville = _ville
                                        else :
                                            rue=get_rue(ent.text,nlp)
                                    elif ent.label_=="CODE_POSTAL" and cp=="" :                                 
                                        cp=get_code_postal(ent.text,nlp)
                                    elif ent.label_=='VILLE' and ville =="":
                                        ville = get_ville(ent.text, nlp)                              
                                if ent.label_ == "CADASTRE" and cadastre_found==False:
                                        cadastre=ent.text
                                        cadastre_found=True
                            if adresse_found == False  and rue!="" and ville!="":
                                # qu'a-t-on trouvé ? 
                                adresse = f'{rue} {cp} {ville}'.strip()
                                if adresse!="" :
                                    logger.debug(f"Adresse= {adresse}") 
                                    adresse_found=True                   
                            if cadastre_found : 
                                logger.debug(f"Cadastre= {cadastre}")                    
            
        else :
            # Rechercher dans le texte s'il reste encore des données à récupérer    
            text,_ = await extract_text_and_table_from_file(file_name=file_path,layout=True,extract_tables=False, extract_text=True)                                      
            if text!=None and text!="" and (num_cu_found==False or adresse_found==False or cadastre_found==False):
                # gestion du text
                results:list[dict[str, str]] =decouper_text_par_titres(text)
                if len(results)> 0 :
                    for result in results : 
                        for key in result :
                            done:bool=False
                            text = result[key]
                            if num_cu=="" and num_cu_found==False:
                                num_cu= get_numero_cu(text)
                                num_cu_found = num_cu!=""   
                                if num_cu_found : done=True                                           
                            if done==False :
                                doc =nlp(text)
                                if adresse_found==False or cadastre_found==False :
                                    rue:str=""
                                    cp:str=""
                                    ville:str=""
                                    for ent in doc.ents : 
                                        if adresse_found == False :
                                            if ent.label_ =="ADRESSE" and rue=="" :                                 
                                                # il faut gérer les adresses qui contiennent la ville :
                                                # 147 rue des Lilas à Aix-en-Provence
                                                is_city_included, _rue,_ville = check_address_with_city(ent.text, nlp)                              
                                                if is_city_included : 
                                                    if _rue!="" : rue =_rue
                                                    if _ville!="" : ville = _ville
                                                else :
                                                    rue=get_rue(ent.text,nlp)
                                            elif ent.label_=="CODE_POSTAL" and cp=="" :                                 
                                                cp=get_code_postal(ent.text,nlp)
                                            elif ent.label_=='VILLE' and ville =="":
                                                ville = get_ville(ent.text, nlp)  
                                        
                                        if ent.label_ == "CADASTRE" and cadastre_found==False:
                                                cadastre=ent.text
                                                cadastre_found=True
                                    if adresse_found == False and rue!="" and ville!="":
                                        # qu'a-t-on trouvé ? 
                                        adresse = f'{rue} {cp} {ville}'.strip()
                                        if adresse!="" :
                                            logger.debug(f"Adresse= {adresse}") 
                                            adresse_found=True                   
                                    if cadastre_found : 
                                        logger.debug(f"Cadastre= {cadastre}")                            
                else :
                    # Pas de découpage par titre :( 
                    text = sanitize_text(text)
                    if num_cu=="" and num_cu_found==False:
                        num_cu= get_numero_cu(text)
                        num_cu_found = num_cu!=""
                    if num_cu_found : 
                        # il faut supprimer le n° de CU trouvé afin de ne plus l'analyser par NLP
                        text=text.replace(num_cu," XXXXXXXXXXXXX ")
                    doc =nlp(text)
                    if adresse_found==False or cadastre_found==False :
                        rue:str=""
                        cp:str=""
                        ville:str=""
                        for ent in doc.ents : 
                            if adresse_found == False :
                                if ent.label_ =="ADRESSE" and rue==""  :                                 
                                    # il faut gérer les adresses qui contiennent la ville :
                                    # 147 rue des Lilas à Aix-en-Provence
                                    is_city_included, _rue,_ville = check_address_with_city(ent.text, nlp)                              
                                    if is_city_included : 
                                        if _rue!="" : rue =_rue
                                        if _ville!="" : ville = _ville
                                    else :
                                        rue=get_rue(ent.text,nlp)
                                elif ent.label_=="CODE_POSTAL" and cp=="":                                 
                                    cp=get_code_postal(ent.text,nlp)
                                elif ent.label_=='VILLE' and ville=="":                            
                                    ville = get_ville(ent.text, nlp)                            
                            if ent.label_ == "CADASTRE" and cadastre_found==False:
                                    cadastre=ent.text
                                    cadastre_found=True
                        if adresse_found == False and rue!="" and ville!="":
                            # qu'a-t-on trouvé ? 
                            adresse = f'{rue} {cp} {ville}'.strip()
                            if adresse!="" :
                                logger.debug(f"Adresse= {adresse}") 
                                adresse_found=True                   
                        if cadastre_found : 
                            logger.debug(f"Cadastre= {cadastre}")                        
            
    return 0,"Ok",num_cu,adresse, cadastre


import logging
import os
from typing import List 
from lxf.ai.text_analysis.default_text_analysis import summarize_chunks
from lxf.settings import get_logging_level
###################################################################

logger = logging.getLogger('Loan2ExtractedData')
fh = logging.FileHandler('./logs/loan2extracteddata.log')
fh.setLevel(get_logging_level())
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.setLevel(get_logging_level())
logger.addHandler(fh)
#################################################################

from lxf.domain.loan import Assurance, Emprunteur, Pret, ProgrammeFinancier, Warranty
from dotenv import dotenv_values
from lxf.services.measure_time import measure_time

from lxf.domain.extracted_data import HIERARCHIE_DOSSIER, Chunk, ChunkMetadata, ExtractedData, ExtractedMetadata
from lxf.domain.predictions import Predictions


def get_metatata(entity_id:str, pret:Pret, p:Predictions)->ExtractedMetadata:
    metadata = ExtractedMetadata()
    metadata.origine = "Ingestion/Offre de prêt"
    metadata.classe = p.BestPrediction
    metadata.classe_confidence= p.BestPredictionConfidence
    metadata.documentId = entity_id

    return metadata

def get_common_section(pret:Pret)->str:
    return f"""
        Information de suivi et de gestion du dossier de prêt
        Numéro de suivi : {pret.num_suivi} 
        Offre : {pret.header_offre} 
        Référence de l'offre : {pret.header_ref} 
        Responsable du suivi de l'offre: {pret.header_responsable_suivi} 
        Numéro Etude : {pret.header_num_etude}                                
        Date d'émission de l'offre : {pret.date_emission_offre} 
    """ 

def get_section_objet_du_financement(pret:Pret)->str :
        
    def get_programme_financier(p:ProgrammeFinancier) -> str:
        msg=""
        for ele in p.plan_financement :  
            msg+=f"""
                  {ele.nature} : {ele.montant.valeur} {ele.montant.devise}
            """
        msg+=f"""
        Montant du Programme : {p.montant_programme.valeur} {p.montant_programme.devise}
        """
        return msg

    msg = f"""
    """
    nb_financement= len (pret.objets_financement)
    for index,financement in enumerate(pret.objets_financement) :
        msg+=f"""  
    Plan de financement avec l'objet et la nature du prêt        
    Financement {index+1}/{nb_financement}   
    Objet du financement : {financement.titre}
    Nature du prêt: {financement.nature}
    Adresse : {financement.adresse} {financement.code_postal} {financement.ville}
    Description complète: 
                {financement.description} 
        """
    msg+=f"""
    PLAN FINANCEMENT :
    {get_programme_financier(pret.programme_financier)}
    """
    msg+="""
    """
    return msg

def get_section_preteur(pret:Pret)->str :
                
    return f"""
        Information sur la banque prêteuse du prêt   
                          
                          
        Nom de la banque : {pret.preteur.raison_sociale}
        Statut de la banque: {pret.preteur.statut_organisation}
        Adresse de la banque : {pret.preteur.adresse} {pret.preteur.code_postal} {pret.preteur.ville}
        SIREN/SIRET : {pret.preteur.siren} {pret.preteur.siret}
    """    

def get_section_emprunteurs(pret:Pret)-> str :
    emprunteurs:List[Emprunteur] = pret.emprunteurs
    section_emprunteurs =f"""
    Liste des emprunteurs et de leur représentant     
    Prêt    
    """
    i=0
    for emprunteur in emprunteurs :
        i+=1
        section_emprunteurs+=f"""
        Emprunteur :
        """
        if emprunteur.est_personne ==False:

            section_emprunteurs+=f"""
            Raison sociale :{emprunteur.raison_sociale}
            Statut : {emprunteur.statut_organisation}
            Siren/Siret: {emprunteur.siren} {emprunteur.siret}
            Adresse : {emprunteur.adresse} {emprunteur.code_postal} {emprunteur.ville}
                """
            for representant in emprunteur.represente_par :
                section_emprunteurs+=f"""
            Représenté par :
                Nom : {representant.civilite} {representant.prenoms} {representant.nom}
                Né(e) le : {representant.date_naissance}
                Situation maritale : {representant.situation_maritale}
                Métier : {representant.metier}
                Adresse : {representant.adresse} {representant.code_postal} {representant.ville}
                    """
        else :
            section_emprunteurs+=f"""
            Nom : {emprunteur.civilite} {emprunteur.prenoms} {emprunteur.nom}
            Né(e) le : {emprunteur.date_naissance}
            Situation maritale : {emprunteur.situation_maritale}
            Métier : {emprunteur.metier}
            Adresse : {emprunteur.adresse} {emprunteur.code_postal} {emprunteur.ville}            
                """
            for representant in emprunteur.represente_par :
                section_emprunteurs+=f"""
            Représenté par :
                Nom : {representant.civilite} {representant.prenoms} {representant.nom}
                Né(e) le : {representant.date_naissance}
                Situation maritale : {representant.situation_maritale}
                Métier : {representant.metier}
                Adresse : {representant.adresse} {representant.code_postal} {representant.ville}
                    """    
     
    return section_emprunteurs + """
    """

def get_list_des_prets(p:Pret)->str :
    msg="""
     Liste des prêts avec leur numéro et montant     
    """
    msg+="_"*70
    msg+="""    
    |{nature:30}|{numero:^18}|{montant:18}|
    """.format(nature="Nature",numero="numero",montant="Montant")
    msg+="="*70
    for ele in p.list_prets :
        msg+="""
    |{nature:30}|{numero:^18}|{valeur:<14} {devise:>3}|
    """.format(nature=ele.nature[:30],numero=ele.numero,valeur=ele.montant.valeur, devise=ele.montant.devise[:3])
    msg+="_"*70
    msg+="""
    
    """
    return msg

def get_caracteristiques_pret(p:Pret)->str :
    msg="""
    Caractéristiques du prêt    
    """
    for caract in p.caracteristiques_prets :
        msg+=f"""
        Prêt: {caract.description} 
        Numéro : {caract.numero_pret}
        Montant total : {caract.montant}
        Domiciliation :
            Débiteur: {caract.domiciliation.debiteur} 
            Compte : {caract.domiciliation.compte}
            Créancier: {caract.domiciliation.creancier}        
        Taux : {caract.taux.description}. Pourcentage: {caract.taux.taux_pourcentage}% Période : {caract.taux.taux_periode}.  Base période: {caract.taux.base_periode}.
        Montant total crédit: {caract.montant_total_du_emprunteurs.cout_total_credit.valeur} {caract.montant_total_du_emprunteurs.cout_total_credit.devise}
        Montant total des intérêts : {caract.montant_total_du_emprunteurs.montant_total_interets.valeur} {caract.montant_total_du_emprunteurs.montant_total_interets.devise}
        Montant total assurances obligatoires : {caract.montant_total_du_emprunteurs.montant_total_assurance_obligatoire.valeur} {caract.montant_total_du_emprunteurs.montant_total_assurance_obligatoire.devise}       
        """
        for frais in caract.montant_total_du_emprunteurs.frais :
            msg+=f"""
            - {frais.nature}: {frais.montant.valeur} {frais.montant.devise}"""            
        msg+=f"""
        Montant total dû emprunteur : {caract.montant_total_du_emprunteurs.montant_total_credit.valeur} {caract.montant_total_du_emprunteurs.montant_total_credit.devise}
        """
        for periode in caract.periodes_amortissement : 
            msg+=f"""
            Type d'échéance : {periode.type_echeance}
            Durée échéances : {periode.duree_echeances}
            Périodicité échéance : {periode.periodicite_echeance}
            Type de taux débiteur: {periode.type_taux_debiteur}
            Taux débiteur : {periode.taux_debiteur_pourcentage}%
            Montant des échéances :
            """
            for montant in periode.montant_echeances:
                msg+=f"""
                {montant.txt_montant}: {montant.montant.valeur} {montant.montant.devise}
                """
            msg+=f"""
            Assurance : {periode.assurance.valeur} {periode.assurance.devise}
            Accessoire: {periode.accessoire.valeur} {periode.accessoire.devise}
            """
    msg+="""
    
    """
    return msg

def get_assurances(p:Pret)->str:
    assurances:list[Assurance] = p.assurances
    msg="""
     Liste des assurances du prêt et des souscripteurs    
    """
    for assurance in assurances : 
        msg+=f"""
        Souscripteur Assuré(e) : {assurance.souscripteur}
            Référence : {assurance.reference}
            Délégation : {assurance.delegation}
            
                Nom de naissance : {assurance.nom_naissance_souscripteur}
                Date de naissance : {assurance.date_naissance_souscripteur}
                Ville de naissance : {assurance.ville_naissance_souscripteur}
            Garanties prises: {assurance.garanties}
            Assureur : {assurance.assureur}
            Type assureur : {assurance.type_assureur}
            Groupe assureur : {assurance.assureur_group}
            Quotité prêt assuré : {assurance.quotite_pret_assure}
            Options choisies : {assurance.options_choisies}
            Hauteur pourcentage : {assurance.hauteur_pourcentage}
            Couverture : {assurance.couverture}
            Tableau des prêts        
        """
        table_prets = assurance.tableau_prets
        msg+="\n"
        nb_rows = len(table_prets.rows)
        for irow, row in enumerate(table_prets.rows):            
            for cell in row.cells :
                msg+=f" {cell.value:40}|"
            # sauter une ligne 
            msg+="\n"            
            if irow+1<nb_rows :
                # une séparation sauf pour la dernière ligne
                for cell in row.cells :
                    msg+="="*42            
                # sauter une ligne 
                msg+="\n"            

        # msg+=f"""
        #     Description complète:

        #         {assurance.description}
            
        #     """
        msg+=f"""
        
        """
    msg+="""
    
    """
    return msg

def get_garanties(p:Pret)->str :
    
    garanties:list[Warranty] = p.garanties
    msg=f"""
    Liste des garanties du prêt 
    """
    for garantie in garanties:
        msg+=f"""
        Garantie : {garantie.label}
        Rang: {garantie.rang}
        Hauteur : {garantie.hauteur}
        Numéro du prêt : {garantie.numero_pret}
        Références cadastrales : {garantie.ref_cadastrales}
            Adresse : {garantie.adresse_immeuble_cadastre}
            Code postal : {garantie.code_poste_immeuble_cadastre}
            Ville : {garantie.ville_immeuble_cadatre}
        Notaire :
            Office : {garantie.office_notarial}
            Addresse : {garantie.adresse_office_notarial}
            Code postal : {garantie.code_postal_office_notarial}
            Ville : {garantie.ville_office_notarial}
        Durée : {garantie.duree} {garantie.duree_unite}
        Durée limitée : {garantie.duree_limitee} {garantie.duree_limitee_unite}
        """
    return msg

def get_clauses_particulieres(p:Pret)->str:
    if len(p.clauses_particulieres)==0 : return "Aucune CLAUSE PARTICULIERE"
    msg="""    
    Liste des clauses particulières du prêt
    """
    for clause in p.clauses_particulieres:
        msg+=f"""
        Titre : {clause.titre}
        Numéro prêt: {clause.numero_pret}
        Texte de la clause :
        {clause.text}

        """
    msg+="""
    -------------
    """
    return msg

def get_tableau_ammortissement(p:Pret)->str:
    if p.tableau_amortissement==None or len(p.tableau_amortissement.rows)==0 :
        return "Pas de tableau d'amortissement"    
    msg="""
    TABLEAU AMORTISSEMENT

    """
    msg+="\n"  
    table_amortissement = p.tableau_amortissement
    nb_rows = len(table_amortissement.rows)
    for irow, row in enumerate(table_amortissement.rows):            
        for cell in row.cells :
            msg+=f" {cell.value:30}|"
        # sauter une ligne 
        msg+="\n"            
        if irow+1<nb_rows :
            # une séparation sauf pour la dernière ligne
            for cell in row.cells :
                msg+="="*32            
            # sauter une ligne 
            msg+="\n"    
    msg+="""
    FIN TABLEAU AMORTISSEMENT
    """
    return msg

def get_summarization(p:Pret)->str:
    """
    Retourne un résumé du prêt
    """
    msg=f"""
    {get_common_section(p)}
    
    {get_section_preteur(p)}
    
    {get_section_emprunteurs(p)}
       
        Plan de financement avec l'objet et la nature du prêt
    """
    nb_financement= len (p.objets_financement)
    for index,financement in enumerate(p.objets_financement) :
        msg+=f"""  
                Financement {index+1}/{nb_financement}   
                Objet du financement : {financement.titre}
                Nature du prêt: {financement.nature}
        """
    msg+=f"""
    
    {get_list_des_prets(p)}
    
    Caractéristiques du prêt
    """
    for caract in p.caracteristiques_prets :
        msg+=f"""
            Prêt: {caract.description} 
            Numéro : {caract.numero_pret}
            Montant total : {caract.montant}
            Domiciliation :
                Débiteur: {caract.domiciliation.debiteur} 
                Compte : {caract.domiciliation.compte}
                Créancier: {caract.domiciliation.creancier}        
            Taux : {caract.taux.description}. Pourcentage: {caract.taux.taux_pourcentage}% Période : {caract.taux.taux_periode}.  Base période: {caract.taux.base_periode}.
            Montant total crédit: {caract.montant_total_du_emprunteurs.cout_total_credit.valeur} {caract.montant_total_du_emprunteurs.cout_total_credit.devise}
            Montant total des intérêts : {caract.montant_total_du_emprunteurs.montant_total_interets.valeur} {caract.montant_total_du_emprunteurs.montant_total_interets.devise}
            Montant total assurances obligatoires : {caract.montant_total_du_emprunteurs.montant_total_assurance_obligatoire.valeur} {caract.montant_total_du_emprunteurs.montant_total_assurance_obligatoire.devise}       
    """    

    msg+=f"""
    
    {get_assurances(p)}
    
    """
    msg+=f"""
    
    {get_garanties(p)}
    
    """
    return msg

@measure_time
def loan_to_extracted_data(pret:Pret,file_name:str, entity_id:str="-1",predictions:Predictions=Predictions())->tuple[int,ExtractedData,str] :
    """
    """
    try :
        
        if pret!=None:
            default_summary=""
            data=ExtractedData()
            nb_chunks:int=11                
            i_chunk:int=1
            data.metadata = get_metatata(entity_id=entity_id, pret=pret,p=predictions)
            chunk =Chunk()
            chunk.metadata=ChunkMetadata()
            chunk.metadata.chunk=i_chunk
            chunk.metadata.chunks=nb_chunks
            chunk.metadata.title="ENTETE"
            chunk.metadata.description="Entête du prêts"
            chunk.metadata.source=file_name
            chunk.page_content=get_common_section(pret)         
            chunk.summary = chunk.page_content
            # Explications
            chunk.explain=f"""
Cet extrait a été obtenu en extrayant l'entête de l'offre de prêt          
                """            
            default_summary+=f"{chunk.metadata.title}\n{chunk.summary}\n"
            data.chunks.append(chunk)
                            
            i_chunk+=1
            data.metadata = get_metatata(entity_id=entity_id, pret=pret,p=predictions)
            chunk =Chunk()
            chunk.metadata=ChunkMetadata()
            chunk.metadata.chunk=i_chunk
            chunk.metadata.chunks=nb_chunks
            chunk.metadata.title="PRETEUR"
            chunk.metadata.source=file_name                    
            chunk.metadata.description="Liste des prêteurs"
            chunk.page_content=get_section_preteur(pret)
            chunk.summary = chunk.page_content
            # Explications
            chunk.explain=f"""
Cet extrait a été obtenu en extrayant la liste des prêteurs          
                """            
            default_summary+=f"{chunk.metadata.title}\n{chunk.summary}\n"            
            data.chunks.append(chunk)
            
            i_chunk+=1
            chunk=Chunk()
            chunk.metadata=ChunkMetadata()
            chunk.metadata.chunk=i_chunk
            chunk.metadata.chunks=nb_chunks
            chunk.metadata.title="EMPRUNTEURS"
            chunk.metadata.description="Liste des emprunteurs"
            chunk.metadata.source=file_name
            chunk.page_content=get_section_emprunteurs(pret)
            # Calculer le résumer et les mots clés
            chunk_summaries = chunk.page_content            
            chunk.summary = chunk_summaries
            # Explications
            chunk.explain=f"""
Cet extrait a été obtenu en extrayant la liste des emprunteurs         
                """            
            default_summary+=f"{chunk.metadata.title}\n{chunk.summary}\n"            
            data.chunks.append(chunk)
            
            i_chunk+=1
            chunk=Chunk()
            chunk.metadata = ChunkMetadata()
            chunk.metadata.chunk=i_chunk
            chunk.metadata.chunks=nb_chunks
            chunk.metadata.title="OBJET des FINANCEMENTS"
            chunk.metadata.description="Objet de tous les financements"
            chunk.metadata.source=file_name
            chunk.page_content = get_section_objet_du_financement(pret)    
            chunk.summary = chunk.page_content
            # Explications
            chunk.explain=f"""
Cet extrait a été obtenu en extrayant l'objet(s) des/du financement(s)        
                """            
            default_summary+=f"{chunk.metadata.title}\n{chunk.summary}\n"            
            data.chunks.append(chunk)

            i_chunk+=1 
            chunk=Chunk()
            chunk.metadata=ChunkMetadata()
            chunk.metadata.chunk= i_chunk
            chunk.metadata.chunks=nb_chunks
            chunk.metadata.title="Liste des prêts"
            chunk.metadata.description="LISTE des PRETS"
            chunk.metadata.source=file_name
            chunk.page_content = get_list_des_prets(pret)         
            chunk.summary = chunk.page_content
            # Explications
            chunk.explain=f"""
Cet extrait a été obtenu en extrayant l'entête de l'offre de prêt          
                """            
            default_summary+=f"{chunk.metadata.title}\n{chunk.summary}\n"            
            data.chunks.append(chunk)

            i_chunk+=1 
            chunk=Chunk()
            chunk.metadata=ChunkMetadata()
            chunk.metadata.chunk= i_chunk
            chunk.metadata.chunks=nb_chunks
            chunk.metadata.title="Caractéristiques"
            chunk.metadata.description="Caractéristiques financiers des PRETS"
            chunk.metadata.source=file_name
            chunk.page_content = get_caracteristiques_pret(pret)       
            chunk.summary = chunk.page_content
            # Explications
            chunk.explain=f"""
Cet extrait a été obtenu en extrayant la liste des prêts         
                """            
            default_summary+=f"{chunk.metadata.title}\n{chunk.summary}\n"            
            data.chunks.append(chunk)

            i_chunk+=1 
            chunk=Chunk()
            chunk.metadata=ChunkMetadata()
            chunk.metadata.chunk= i_chunk
            chunk.metadata.chunks=nb_chunks
            chunk.metadata.title="Assurances"
            chunk.metadata.description="Liste des assurances souscrites"
            chunk.metadata.source=file_name
            chunk.page_content = get_assurances(pret)          
            chunk.summary = chunk.page_content
            # Explications
            chunk.explain=f"""
Cet extrait a été obtenu en extrayant la liste des assurances souscriptes par les emprunteurs         
                """            
            default_summary+=f"{chunk.metadata.title}\n{chunk.summary}\n"            
            data.chunks.append(chunk)

            i_chunk+=1 
            chunk=Chunk()
            chunk.metadata=ChunkMetadata()
            chunk.metadata.chunk= i_chunk
            chunk.metadata.chunks=nb_chunks
            chunk.metadata.title="Garanties"
            chunk.metadata.description="Liste des garanties prises"
            chunk.metadata.source=file_name
            chunk.page_content = get_garanties(pret)         
            chunk.summary = chunk.page_content
            # Explications
            chunk.explain=f"""
Cet extrait a été obtenu en extrayant la liste des emprunteurs       
                """            
            default_summary+=f"{chunk.metadata.title}\n{chunk.summary}\n"            
            data.chunks.append(chunk)

            i_chunk+=1 
            chunk=Chunk()
            chunk.metadata=ChunkMetadata()
            chunk.metadata.chunk= i_chunk
            chunk.metadata.chunks=nb_chunks
            chunk.metadata.title="Clauses particulières"
            chunk.metadata.description="Liste des clause particuluères"
            chunk.metadata.source=file_name
            chunk.page_content = get_clauses_particulieres(pret)
            # Calculer le résumer et les mots clés
            if len(chunk.page_content)>30 :                 
                tmp = summarize_chunks(chunk.page_content,summary_max_length=512) 
                chunk_summaries = "".join(tmp)
            else :
                chunk_summaries = chunk.page_content            
            chunk.summary = chunk_summaries
            # Explications
            chunk.explain=f"""
Cet extrait a été obtenu en extrayant la liste des clauses particulières          
                """            
            default_summary+=f"{chunk.metadata.title}\n{chunk.summary}\n"            
            data.chunks.append(chunk)

            i_chunk+=1 
            chunk=Chunk()
            chunk.metadata=ChunkMetadata()
            chunk.metadata.chunk= i_chunk
            chunk.metadata.chunks=nb_chunks
            chunk.metadata.title="Tableau d'amortissement"
            chunk.metadata.description="Tableau des amortissement"
            chunk.metadata.source=file_name
            chunk.page_content = get_tableau_ammortissement(pret)          
            chunk.summary = chunk.page_content
            # Explications
            chunk.explain=f"""
Cet extrait a été obtenu en extrayant le tableau d'amortissement         
                """            
            # pour le résumé, on ne prend pas le tableau d'amortissement
            #default_summary+=f"{chunk.metadata.title}\n{chunk.summary}\n"
            data.chunks.append(chunk)
            
            i_chunk+=1 
            chunk=Chunk()
            chunk.metadata=ChunkMetadata()
            chunk.metadata.chunk= i_chunk
            chunk.metadata.chunks=nb_chunks
            chunk.metadata.title="Résumé"
            chunk.metadata.description="Résumé de l'offre de prêt"
            chunk.metadata.hierarchie=HIERARCHIE_DOSSIER
            chunk.metadata.source=file_name
            chunk.page_content =default_summary
            chunk.summary =  get_summarization(p=pret)     
            # Explications
            chunk.explain=f"""
Cet extrait a été obtenu en extrayant toutes les parties caractéristiques de l'offre de prêt         
                """            
            data.chunks.append(chunk)             
                    
            data.keywords.append("PRET")
            data.keywords.append("Bancaire")
            data.keywords.append("BPACA")
            data.keywords.append("PTZ")    
            data.keywords.append("GARANTIES")
            data.keywords.append("ASSURANCES")         
            
            return 0,data ,"sucess"  
    except Exception as ex: 
        err_msg=f"Exception dans Loan to_ExtractedData:\n{ex}"
        logger.critical(err_msg,stack_info=True)
        return -1,None, err_msg


import json
import logging
import os
import re

from lxf.ai.agents.interviewer import Interviewer
from lxf.ai.ocr.pdf_utilities import remove_header_footer, remove_highligth
from lxf.domain.loan import Emprunteur, Montant, ObjetFinancement, Personne, Preteur
from lxf.extractors.finance.loans.loan_extractor import get_text_and_tables_from_pdf
from lxf.domain.loan_proposal import ConditionsFinancieres, Garantie, GarantieFinanciereAchevement, LoanProposal, ModalitesFonctionnement
from lxf.ai.ocr.ocr_pdf import do_ocr
from lxf.domain.tables import lxfTable
from lxf.services.try_safe import try_safe_execute, try_safe_execute_async 

SETLEVEL = logging.DEBUG
logger = logging.getLogger('Loan Proposal Extractor')
fh = logging.FileHandler('./logs/loan_proposal_extractor.log')
fh.setLevel(SETLEVEL)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.setLevel(SETLEVEL)
logger.addHandler(fh)

def sanitize_logo(text:str)->str :
    """
    Docstring for sanitize_logo
    
    :param text: Description
    :type text: str
    :return: Description
    :rtype: str
    """
    regex=r"\+X, BOURGOGNE FRANCHE-COMTE"
    subst=""
    result = re.sub(regex,subst,text,0,re.MULTILINE|re.DOTALL)
    return result

def sanitize_footer(text:str)->str:
    """
    Docstring for sanitize_footer
    
    :param text: Description
    :type text: str
    :return: Description
    :rtype: str
    """
    regex=r"Les personnes c.ncern.es.*?bpbfc\.(b|h)anquepopulaire\.fr"
    subst=""
    text = re.sub(regex,subst,text,0,re.MULTILINE|re.DOTALL)
    # second footer paragraph
    regex = r"BPBFC Soci.t..*?N° ORIAS Courtier Assurances : [0-9]* *\."
    result = re.sub(regex,subst,text,0,re.MULTILINE|re.DOTALL)
    return result    

def sanitize_bullet(text:str)->str :
    """
    Docstring for sanitize_bullet
    
    :param text: Description
    :type text: str
    :return: Description
    :rtype: str
    """
    regex=r"^e "
    subst="* "
    result = re.sub(regex,subst,text,0,re.MULTILINE|re.DOTALL)
    return result 

def sanitize_miscelanous(text:str)->str:
    """
    Docstring for sanitize_miscelanous
    
    :param text: Description
    :type text: str
    :return: Description
    :rtype: str
    """
    subst=""
    regex=r"\\"
    result = re.sub(regex, subst, text,0,re.MULTILINE|re.DOTALL)
    regex=r"swift.*BOURGOGNE Franche-COMTE\n"
    result = re.sub(regex,subst,result,0,re.MULTILINE|re.DOTALL|re.IGNORECASE)
    regex=r"( [A-Za-z0-9,éèù'/àç] +[A-Za-z0-9éèà'ù/,ç] )"
    result = re.sub(regex,subst,result,0,re.MULTILINE|re.DOTALL|re.IGNORECASE)
    return result

def extract_json(text:str)->tuple[int,str|None] :
    """
    Docstring for extract_json
    
    :param text: Description
    :type text: str
    :return: Description
    :rtype: tuple [int,str | None]
    """
    regex=r"(?P<json>{.*})"
    matches=re.findall(regex,text,re.DOTALL|re.MULTILINE)
    if matches !=None and len(matches)>0:
        json_str=matches[0]     
        return 0,json_str
    logger.warning(f"Extraction d'un JSON impossible depuis la chaine:\n{text}")
    return -1, None
    

class LoanProposalDataExtractor :
    """
    Docstring for LoanProposalDataExtractor
    """
    QUESTION="question"
    EXTRACTEUR="extracteur"
    
    def __init__(self,filename:str, base_url:str, model_instruct:str,openapi_compatibility=False) -> None:
        """
        Ctor         
        filename : the file name 
        """
        if os.path.exists(filename) :
            self.filename=filename
            self.base_url=base_url
            self.model_instruct=model_instruct
            self.openapi_compatibility = openapi_compatibility
        else :
            raise NameError(f"{filename} non trouvé")    

    async def ask(self,interviewer:Interviewer, question:str,proposition)->tuple[int,str]:
        """
        Docstring for ask
        
        :param question: La question à poser
        :type question: str
        :param propositon: Texte de la proposition sur laquelle on pose une question
        :return: Le code erreur et la réponse si le code d'erreur = 0 (OK)
        :rtype: tuple[int, str]
        """
        full_question=f"""{question}\n - Utilise le texte ci-dessous pour réondre à la question \n{proposition}"""
        error, response = await interviewer.invoke_async(full_question)
        return error, response
    
    def extract_preteur(self,json_str:str, loan_proposal:LoanProposal)->tuple[int,LoanProposal] :
        """
        Docstring for extract_preteur
        
        :param self: Description
        :param json_str: Description
        :type json_str: str
        :return: Description
        :rtype: tuple int, LoanProposal
        """
        try:
            err,json_str=extract_json(json_str)
            if err!=0 or json_str == None : return -3, loan_proposal
            data = json.loads(json_str)
            if data!=None : 
                preteur:Preteur=Preteur()
                preteur.raison_sociale = data.get('preteur',"")
                preteur.adresse = data.get("adresse","")
                preteur.code_postal=data.get("code_postal","")
                preteur.ville = data.get("ville","")
                preteur.source=data.get('source',"")
                loan_proposal.preteur = preteur
                return 0,loan_proposal
            else :
                return -1 , loan_proposal
        except Exception as ex:
            logger.exception(f"Exception pendant l'extraction des donnees:\n{ex}\nJSON_STR:\n{json_str} ")
            return -2 , loan_proposal         

    def extract_emprunteur(self, json_str:str, loan_proposal:LoanProposal)->tuple[int,LoanProposal]:
        """
        Docstring for extract_emprunteur
        
        :param self: Description
        :param json_str: Description
        :return: Description
        :rtype: tuple[int, Emprunteur]
        """
        try :
            err,json_str=extract_json(json_str)
            if err!=0 or json_str == None : return -3, loan_proposal         
            data = json.loads(json_str)
            if data!=None : 
                emprunteurs=data.get('emprunteurs',None)
                source=data.get("source",None)
                rs = data.get("raison_sociale",None)
                if rs!=None and rs!="" :
                    # Personne morale
                    emprunteur:Emprunteur=Emprunteur()
                    emprunteur.raison_sociale=rs
                    emprunteur.est_personne=False
                    emprunteur.source=source
                    emprunteur.represente_par =[]                
                    for emp in emprunteurs: 
                        pers:Personne=Personne()
                        pers.nom = emp.get("nom","")
                        pers.prenoms=emp.get("prenom","")
                        pers.est_personne=True
                        emprunteur.represente_par.append(pers)
                    loan_proposal.emprunteurs.append(emprunteur)
                else :
                    # Personnes physiques
                    for emp in emprunteurs :
                        emprunteur:Emprunteur=Emprunteur()                   
                        emprunteur.nom = emp["nom"]
                        emprunteur.prenoms = emp["prenom"]
                        emprunteur.est_personne=True
                        emprunteur.source=source
                        loan_proposal.emprunteurs.append(emprunteur)
                return 0, loan_proposal
            else:
                return -1, loan_proposal                    
        except Exception as ex:
            logger.exception(f"Exception pendant l'extraction des donnees:\n{ex}\nJSON_STR:\n{json_str} ")
            return -2 , loan_proposal

    def extract_header(self, json_str:str, loan_proposal:LoanProposal)->tuple[int,LoanProposal]:
        """
        Docstring for extract_header        
        :param self: Description
        :param json_str: Description
        :param loan_proposal: Description
        :type loan_proposal: LoanProposal
        :return: Description
        :rtype: tuple[int, LoanProposal]
        """
        try:
            err,json_str=extract_json(json_str)
            if err!=0 or json_str == None : return -3, loan_proposal            
            data = json.loads(json_str)
            if data!=None :
                loan_proposal.header_ref = data.get("reference","")
                loan_proposal.header_offre = data.get("titre","")
                loan_proposal.header_responsable_suivi=data.get("responsable","")
                loan_proposal.header_num_etude =data.get("numéro","")
                loan_proposal.date_emission_offre = data.get("date_offre","")
                return 0, loan_proposal 
            else :
                return -1, loan_proposal  
        except Exception as ex:
            logger.exception(f"Exception pendant l'extraction des donnees:\n{ex}\nJSON_STR:\n{json_str} ")
            return -2 , loan_proposal              
      
    def extract_objet_financement(self, json_str:str,loan_proposal:LoanProposal)->tuple[int,LoanProposal]:
        """
        Docstring for extract_objet_financement
        
        :param self: Description
        :param json_str: Description
        :type json_str: str
        :param loan_proposal: Description
        :type loan_proposal: LoanProposal
        :return: Description
        :rtype: tuple[int, LoanProposal]
        """
        try:
            err,json_str=extract_json(json_str)
            if err!=0 or json_str == None : return -3, loan_proposal            
            data = json.loads(json_str)
            if data !=None :
                objet:ObjetFinancement = ObjetFinancement()
                objet.titre = data.get("titre","")
                objet.description = data.get("description","")
                objet.adresse = data.get("adresse","")
                objet.code_postal=data.get("code_postal","")
                objet.ville = data.get("ville","")
                objet.usage =data.get("usage","")
                objet.nature=data.get("nature","")
                objet.deblocage_plusieurs_phases = data.get("deblocage_plusieurs_phases","False")
                objet.phase_1=data.get("phase_1","")
                objet.phase_2=data.get("phase_2","")
                objet.condition_credit_phase_1 = data.get("condition_credit_phase_1","")
                objet.condition_credit_phase_2 = data.get("condition_credit_phase_2","")
                objet.source = data.get("source","")                
                loan_proposal.objets_financement.append(objet)
                return 0, loan_proposal
        except Exception as ex:
            logger.error(f"Exception : {ex}")        
        return -1, loan_proposal    
    def extract_conditions_financieres(self,json_str:str,loan_proposal:LoanProposal)->tuple[int,LoanProposal]:
        """
        Docstring for extract_conditions_financiers
        
        :param self: Description
        :param json_str: Description
        :type json_str: str
        :param loan_proposal: Description
        :type loan_proposal: LoanProposal
        :return: Description
        :rtype: tuple[int, LoanProposal]
        """
        try:
            err,json_str=extract_json(json_str)
            if err!=0 or json_str == None : return -3, loan_proposal            
            data = json.loads(json_str)
            if data !=None :
                cf:ConditionsFinancieres= loan_proposal.conditions_financieres
                cf.source = data.get("source","")
                cf.commission_engagement =data.get("commission_engagement","")
                m:Montant=Montant()
                frais_dossier=data.get("frais_dossier",m)
                m.devise=frais_dossier.get("devise","EUR").replace("€","EUR")
                m.valeur = frais_dossier.get("valeur",0.0)
                cf.frais_dossier=m
                cf.is_taux_variable = data.get("is_taux_variable")
                cf.calcul_taux_variable=data.get("calcul_taux_variable","")
                cf.periodicite_commission=data.get("periodicite_commission","")
                cf.depassement_autorisation=data.get("depassement_autorisation","")
                cf.methode_calcul_commission_engagement=data.get("methode_calcul_commission_engagement","")
                return 0, loan_proposal
            return -1, loan_proposal
        except Exception as ex:
            logger.exception(f"Exception pendant l'extraction des donnees:\n{ex}\nJSON_STR:\n{json_str} ")
            return -2 , loan_proposal
    def extract_gfa(self, json_str:str,loan_proposal:LoanProposal)->tuple[int,LoanProposal]:
        """
            :param self: Description
            :param json_str: Description
            :type json_str: str
            :param loan_proposal: Description
            :type loan_proposal: LoanProposal
            :return: Description
            :rtype: tuple[int, LoanProposal]
            """
        try :
            err,json_str=extract_json(json_str)
            if err!=0 or json_str == None : return -3, loan_proposal            
            data = json.loads(json_str)
            if data !=None :
                source = data.get("source","")
                return 0, loan_proposal
            return -1, loan_proposal    
        except Exception as ex:
            logger.exception(f"Exception pendant l'extraction des donnees:\n{ex}\nJSON_STR:\n{json_str} ")
            return -2 , loan_proposal
        
    def extract_garanties_et_conditions(self, json_str:str,loan_proposal:LoanProposal)->tuple[int,LoanProposal]:
        """
            :param self: Description
            :param json_str: Description
            :type json_str: str
            :param loan_proposal: Description
            :type loan_proposal: LoanProposal
            :return: Description
            :rtype: tuple[int, LoanProposal]
            """
        try :
            err,json_str=extract_json(json_str)
            if err!=0 or json_str == None : return -3, loan_proposal            
            data = json.loads(json_str)
            if data !=None :
                source = data.get("source","")
                loan_proposal.condition_financiere_achevement.source=source
                loan_proposal.condition_credit_accompagnement.source=source            
                garanties=data.get("garanties",[])
                if garanties !=[] :
                    for garantie in garanties : 
                        g:Garantie = Garantie()
                        g.label = garantie.get("titre","")
                        m:Montant=Montant()
                        hauteur = garantie.get("hauteur",m)
                        m.devise=hauteur.get("devise","EUR").replace("€","EUR")
                        m.valeur=hauteur.get("valeur",0.0)
                        g.hauteur=m
                        g.rang=garantie.get("rang","")
                        g.rang = g.rang.replace("%","").strip()
                        g.cautionnaires= garantie.get("cautionnaires",[])
                        g.engagements=garantie.get("engagements",[])                       
                        loan_proposal.garanties.append(g)         
                conditions_credit_accompagnement = data.get("conditions_credit_accompagnement",[])
                if conditions_credit_accompagnement!=[] :
                    for condition_ca in conditions_credit_accompagnement :
                        loan_proposal.condition_credit_accompagnement.conditions.append(condition_ca)
                condition_financiere_achevement = data.get("condition_financiere_achevement",[])
                if condition_financiere_achevement !=[] :
                    for condition_fa in condition_financiere_achevement : 
                        loan_proposal.condition_financiere_achevement.conditions.append(condition_fa)                                                
                return 0, loan_proposal
            return -1, loan_proposal    
        except Exception as ex:
            logger.exception(f"Exception pendant l'extraction des donnees:\n{ex}\nJSON_STR:\n{json_str} ")
            return -2 , loan_proposal
        
    def extract_modalites_fonctionnement(self,json_str:str,loan_proposal:LoanProposal)->tuple[int,LoanProposal] :
        """
        Docstring for extract_modalites_fonctionnement
        
        :param self: Description
        :param json_str: Description
        :type json_str: str
        :param loan_proposal: Description
        :type loan_proposal: LoanProposal
        """
        try:
            err,json_str=extract_json(json_str)
            if err!=0 or json_str == None : return -3, loan_proposal            
            data = json.loads(json_str)
            if data !=None :
                mf:ModalitesFonctionnement=loan_proposal.modalites_fonctionnement 
                mf.modalites = data.get("modalites_fonctionnement",[])      
                mf.date_limite_validite = data.get("date_limite_validite","")      
                mf.source = data.get("source","")
                return 0, loan_proposal
            return -1, loan_proposal        
        except Exception as ex:
            logger.exception(f"Exception pendant l'extraction des donnees:\n{ex}\nJSON_STR:\n{json_str} ")
            return -2 , loan_proposal
                        
    async def extract_data(self) ->LoanProposal :
        """
        Extrait les données de la proposition
        """
        proposal:LoanProposal=LoanProposal()
        found_tables=list[lxfTable]
        required_ocr_cleanup=False
        logger.info(f"# Debut traitement {self.filename}")
        text,found_tables = await try_safe_execute_async(logger,get_text_and_tables_from_pdf,filename=self.filename)
        # Eliminer les textes inférieurs à 200 caractères
        if len(text)<=200 : text=""
        if text == None or text=="" : 
            # try with OCR  
            # Remove HighLight and convert to b&w 
            logger.info("Pretraitement image")
            prepared_pdf= self.filename.replace('.pdf','_prepared.pdf')    
            logger.info("Suppression High ligth")        
            err, error_message = remove_highligth(self.filename, prepared_pdf)    
            if err<0 :
                logger.error(f"Remove hightligth a echoue : {error_message}")
                return proposal    
            logger.info("Suppression Entete et Pied de page")   
            removed_header_footer_pdf = prepared_pdf.replace('_prepared.pdf','_removed_header_footer.pdf')
            err, error_message = remove_header_footer(prepared_pdf,removed_header_footer_pdf)
            os.remove(prepared_pdf)
            if err<0 :
                logger.error(f"Remove Header & Footer failed {error_message}")
                return proposal
            output_ocr_pdf= removed_header_footer_pdf.replace('_removed_header_footer.pdf','_ocr.pdf')
            logger.info("OCR")
            result = do_ocr(input_pdf=removed_header_footer_pdf, output_pdf=output_ocr_pdf)  
            os.remove(removed_header_footer_pdf)
            if os.path.exists(output_ocr_pdf):
                ## Todo Processing the data    
                required_ocr_cleanup=True        
                text,found_tables = await try_safe_execute_async(logger,get_text_and_tables_from_pdf,filename=output_ocr_pdf)
                os.remove(output_ocr_pdf)
        if text!=None :    
            text=sanitize_miscelanous(sanitize_bullet(sanitize_footer(sanitize_logo(text))) )   
            interviewer:Interviewer=Interviewer(
                                                base_url=self.base_url,
                                                system_prompt="tu es un assistant français spécialisé en gestion de prêts bancaires",
                                                ctx=100*1024,
                                                model=self.model_instruct,
                                                openapi_compatibility=self.openapi_compatibility
                                                )              
            if required_ocr_cleanup :       
                nettoyage=f"""
    Voici un texte issue d'un OCR, ta mission consiste à corriger les mots mal orthographiés ou incomplets. 
    Voici le texte à traiter: 
    {text}
    """         
                logger.info("Nettoyage du texte par l'IA")
                error, corrected_text = await interviewer.invoke_async(nettoyage)
                # print the corrected text
                if error==0: 
                    # print(corrected_text)                
                    # print("\n"*3)
                    text=corrected_text
                else :
                    logger.warning(f"La correction du texte par l'IA a echouee. Le texte sera traite sans correction")
            proposal.source.append(text)
            # start the interview 
            requests=[
                {self.QUESTION:
                """
                Qui est l'établissement prêteur ? quelle est l'adresse du siège social? 
                # Consignes: 
                - Donne ta réponse dans un json avec impérativement les éléments suivants: 
                    "preteur" : doit contenir l'identification du prêteur,
                    "adresse": doit contenir l'adresse du siège social,
                    "code_postal": doit contenir le code postal, 
                    "ville": doit contenir la ville , 
                    "source" : la ou les parties du texte qui ont permis d'identifier les emprunteurs sous la forme de liste de string.
                - Donne uniquement le json brute sans aucune autre information.
                """,self.EXTRACTEUR:self.extract_preteur},
                {self.QUESTION:
                """
                 Trouve tous les emprunteurs et la raison sociale de l'entreprise, si elle est précisiée, à qui est adressé cette proposition.
                # Consignes: 
                - Ignore tous les contacts de la banque y compris dans la signature après le nom de la banque
                - Les emprunteurs peuvent être une personne ou une entreprise représentée par une ou plusieurs personnes.
                - Cette information se trouve par ordre de priorité :  
                    1- Après le texte suivant : 'A l'attention de ...' ou 'À l'atention de ...' 
                    2- Dans le cas d'une entreprise, ajoute aux données précédentes la raison sociale qui apparait pour la signature de la proposition. Généralement après le texte : Pour la ...                       
                - Donne ta réponse dans un json avec impérativement les éléments suivants:
                    - "raison_sociale": doit contenir, si l'information est présente, la raison sociale de l'entreprise. Sinon la chaine.
                    - "emprunteurs" : doit contenir la liste des emprunteurs. Chaque emprunteur doit contenir le nom et le prémon. 
                    - "source": la ou les parties du texte qui ont permis d'identifier les emprunteurs sous la forme de liste de string.
                - Donne uniquement le json brute sans aucune autre information.                 
                 """,self.EXTRACTEUR:self.extract_emprunteur},
                {self.QUESTION:
                """
                 Quels sont les informations suivantes :
                 - La **référence** de l'offre,
                 - le **titre** de l'offre,
                 - le **responsable** de l'offre,
                 - le **numéro** de l'offre,
                 - la **date** de l'offre.
                 # Consignes :
                - Donne ta réponse dans un json avec impérativement les éléments suivants:
                    - 'reference',
                    - 'titre',
                    - 'responsable':
                    - 'numéro' 
                    - 'date_offre' 
                    - "source": la ou les parties du texte qui ont permis d'identifier les emprunteurs sous la forme de liste de string.
                - Donne uniquement le json brute sans aucune autre information. 
                 """,self.EXTRACTEUR:self.extract_header},
                {self.QUESTION:
                    """
                    Répond à tous les questions ci-dessous, en suivant les consignes données
                       - Quelle est l'objet principal et l'adresse du bien du financement?
                       - Quel besoin va couvrir le financement ?
                       - A quel usage est destiné le financement ?
                       - Existe-il un déblocage en 2 phases ou 2 temps? 
                       - Si Déblocage en 2 phases ou 2 temps: 
                            - Donne moi la description détaillée de la Phase 1. 
                            - Donne moi la description détaillée de la Phase 2.
                            - Donne moi les conditions Crédit de la phase 1.
                            - Donne moi les conditions Crédit de la phase 2
                        # Consignes :
                            - Donne ta réponse dans un json avec impérativement les éléments suivants:
                                - "description" : doit contenir l'objet du fiancement,
                                - "titre": doit contenir le besoin du financement, 
                                - "adresse" : doit contenir l'adresse du financement
                                - "code_postal": doit contenir le code postal du bien financé,
                                - "ville" : doit contenir la ville du bien financé,
                                - "nature" : doit contenir la nature du bien financé,
                                - "usage": doit contenir l'usage du bien financé,
                                - "deblocage_plusieurs_phases": True s'il existe un déblocage en 2 phases ou 2 temps, False sinon.
                                - si déblocage en 2 phases ou 2 temps :
                                    - "phase_1": doit contenir la description détaillée de la phase 1.
                                    - "phase_2": doit contenir la description détaillée de la phase 2.
                                    - "condition_credit_phase_1": doit contenir la description des conditions crédit de la phase 1.
                                    - "condition_credit_phase_2": doit contenir la description des conditions crédit de la phase 2.
                                - "source": la ou les parties du texte qui ont permis d'identifier les emprunteurs sous la forme de liste de string.
                            - Donne uniquement le json brute sans aucune autre information.
                    """,
                    self.EXTRACTEUR:self.extract_objet_financement},
                {self.QUESTION:
                    """
                    Ta mission consiste à extraire les conditions financières:
                    - Vérifier s'il s'agit d'un taux fixe ou variable, 
                        - S'il s'agit d'un taux variable:
                            - Le calcul indexation du taux ,
                            - Le taux du jour
                            - Toutes les informations concernant ce taux
                    - Que se passe-t-il en cas de dépassement de l'autorisation, 
                    - la périodicité à laquelle sera perçu la commission d'engagement,
                    - le montant des frais de dossier contenu dans le paragraphe des conditions financiers,
                    - le pourcentage annuel de la commission d'engagement   
                    - La méthode de calcul de la commission d'engagement           
                    # Consignes : 
                        - Donne ta réponse dans un json avec impérativement les éléments suivants:
                            - "commission_engagement" : doit contenir le pourcentage annuel ,
                            - "depassement_autorisation": doit contenir se qui ce passe en cas de dépassement de l'aurotisation, 
                            - "methode_calcul_commission_engagement": doit contenir la méthode de calcul de la commission d'engagement,
                            - "is_taux_variable": True si le taux est variable, False sinon,                         
                            - "calcul_taux_variable": si le taux de commission est variable, doit contenir les modalités détaillées du calcule du taux variable, 
                            - "periodicite_commission": doit contenir périodicité à laquelle sera perçue la commission d'engagement,
                            - "frais_dossier": doit contenir le montant des frais de dossier sous la forme {"valeur":<montant des frais de dossier>, "devise":<devise monétaire du montant des frais de dossier>}, si pas de montant trouvé , utilise {"valeur":0.0,"devise":"EUR"}
                            - "source": la ou les parties du texte qui ont permis d'identifier les emprunteurs sous la forme de liste de string.
                        - Donne uniquement le json brute sans aucune autre information.                    
                    """,
                    self.EXTRACTEUR:self.extract_conditions_financieres },                    
                {self.QUESTION:
                    """
                    Ta mission consiste à extraire les garanties et conditions de mise en place: 
                    - Garanties: 
                        - Donne la liste de toutes les garanties. 
                            Pour chaque garantie :
                            - Dans le cas d'une hypothèque, trouve le rang et la hauteur en euros (€) ainsi que l'objet de l'hypothèque?
                            - Dans le cas d'une caution, donne tous les détails de la caution en particulier la hauteur en euros(€), la partie cautionnaire et qui recueille cette caution.
                            - Quels sont engagements qui ont été pris et par qui ? 
                            - Existe-t-il des nantissements et si oui lesquels? 
                    - Conditions :
                        - Donne la liste de toutes les condiions :
                            - Distingue une liste pour les conditions Crédit d'accompagnement et des Conditions garantie financière d'achèvement(autrement appelées: GFA). 
                    # Consignes :
                    - Donne ta réponse dans un json avec impérativement les éléments suivants:
                        - "garanties" : doit contenir la liste de toutes les garanties. 
                            - Chaque "garantie" : doit contenir 
                                - "titre": le titre de la garantie ou de la caution,
                                - "hauteur" : le montant en euros(€), sous la forme :{"valeur":<montant de la hauteur>,"devise":<devise monétaire du montant de la hauteur>}, si pas de montant trouvé , utilise {"valeur":0.0,"devise":"EUR"}
                                - "rang" : s'il existe le niveau du rang, la chaine vide sinon.
                                - "cautionnaires": s'il existe une caution,  la liste des toutes personnes physiques ou morales qui se sont portées caution.
                                - "engagements": la liste de tous les engagements pris. 
                        - "conditions_credit_accompagnement": la liste des toutes les conditions Crédit d'accompagnement.
                        - "condition_financiere_achevement" : La liste des toutes les conditions de garantie financière d'achèvement ou GFA .  
                        - "source": la ou les parties du texte qui ont permis d'identifier les emprunteurs sous la forme de liste de string.
                    - Donne uniquement le json brute sans aucune autre information.                         
                    """,
                    self.EXTRACTEUR:self.extract_garanties_et_conditions},
                {self.QUESTION:
                    """
                    Ta mission consiste à extraire toutes les modalités de fonctionnent de la proposition de crédit :
                    - Modalités de fonctionnement :
                        - Donne la liste de toutes les modalités de fonctionnement. 
                    - Date limite de validité de l'accord. 
                    # Consignes :
                    - Donne ta réponse dans un json avec impérativement les éléments suivants:
                        - "modalites_fonctionnement": la liste des toutes les modalités de fonctionnement.
                        - "date_limite_validite": La date limite de validité de l'accord. Date en texte et respectant le format dd/mm/Y  
                        - "source": la ou les parties du texte qui ont permis d'identifier les emprunteurs sous la forme de liste de string.
                    - Donne uniquement le json brute sans aucune autre information.                         
                    """,
                    self.EXTRACTEUR:self.extract_modalites_fonctionnement },
                    # "Quelles sont les modalités de fonctionnement",
                    # "Quelles sont les conditions générales"
            ]
            logger.info(f"Extraction des donnees par l'IA en {len(requests)} etapes")
            for _i,request in enumerate(requests) :
                logger.info(f"Interrogation des donnees etape {_i+1}")
                error, response = await self.ask(interviewer,request[self.QUESTION],text)
                extracteur = request.get(self.EXTRACTEUR,None)
                if extracteur !=None :
                    if error==0 : 
                        # pas d'erreur
                        logger.info("Traitement des donnees extraites")
                        error, proposal = try_safe_execute(logger,extracteur,json_str=response, loan_proposal=proposal)
                        if error!=0 :
                            logger.error(f"Extraction des donnees impossible(error={error}). Réponse:\n{response}")
                        else: 
                            logger.info(f"Extraction des donnees etape {_i+1} reussie")
                    else :
                        logger.warning(f"La question {request.get(self.QUESTION,'')} n'a pu etre repondue")
                else : 
                    logger.error(f"Erreur Extracteur, introuvable !\nQuestion:\n{request.get(self.QUESTION,"")}")
        else :
            logger.error("Aucun texte n'a pu etre extrait ! ")
        logger.info(f"# Fin Traitement")
        return proposal


import os
import cv2
from pdf2image import convert_from_path, convert_from_bytes
from pdf2image.exceptions import (
    PDFInfoNotInstalledError,
    PDFPageCountError,
    PDFSyntaxError
)
import img2pdf

import glob

def empty_folder(path:str) :
    files = glob.glob(path)
    for f in files:
        os.remove(f)

def remove_highligth(source:str, target:str)-> tuple[int,str] :
    """
    Docstring for remove highligh from images of PDF and return a PDF images with the b&w converted images 
    
    :param source: Description
    :type source: str
    :param target: Description
    :type target: str
    :return: Description
    :rtype: tuple[int, str]
    """
    if os.path.exists(source) is False :
        return -1 , f"{source} est introuvable"
    
    images = convert_from_path(source)
    if images == None : 
        return -2 , f"Aucune image extraite du fichier {source}"
    if os.path.exists("./data/temp") is False : 
        os.makedirs("./data/temp")
        
    imgs=[]
    for i in range(len(images)):    
    ## Save image to file
            img_path=f"./data/temp/page_{i}.jpeg"
            images[i].save(img_path,'JPEG')          
            image = cv2.imread(img_path)
            # Convert to Gray
            gray = cv2.cvtColor(image,cv2.COLOR_BGR2GRAY)            
            gray_path= f"./data/temp/gray_page_{i}.png"
            cv2.imwrite(gray_path,gray)
            # display(gray_path)
            min=155
            max=255
            source=gray
            thresh, img_bw = cv2.threshold(source,min,max,cv2.THRESH_BINARY)
            black_white_path =f"./data/temp/bw_page{i}.png"
            cv2.imwrite(black_white_path, img_bw)
            #display(black_white_path)
            imgs.append(black_white_path)            
            os.remove(img_path)
            
    # Save all converted images in a pdf
    # opening from filename
    # specify paper size (A4)
    a4inpt = (img2pdf.mm_to_pt(210),img2pdf.mm_to_pt(297))
    layout_A4 = img2pdf.get_layout_fun(a4inpt)
    with open(target,"wb+") as f:
        f.write(img2pdf.convert(imgs,layout_fun=layout_A4))
    # remove all tempory file
    empty_folder("./data/temp/*.png")
    if os.path.exists(target) is True :
        return 0, target
    return -3, f"Le fichier {target} n'a pu être créé"

def remove_header_footer(source:str,target:str)->tuple[int,str] :
    """
    Docstring for remove_header_footer
    
    :param source: Description
    :type source: str
    :param target: Description
    :type target: str
    :return: Description
    :rtype: tuple[int, str]
    """
    if os.path.exists(source) is False :
        return -1 , f"{source} est introuvable"
    
    images = convert_from_path(source)
    if images == None : 
        return -2 , f"Aucune image extraite du fichier {source}"
    if os.path.exists("./data/temp") is False : 
        os.makedirs("./data/temp")
        
    imgs=[]
    for i in range(len(images)): 
            img_path=f"./data/temp/page_{i}.jpeg"
            images[i].save(img_path,'JPEG')          
            image = cv2.imread(img_path)        
            # Convert to Gray
            gray = cv2.cvtColor(image,cv2.COLOR_BGR2GRAY) 
            # blur the image
            kernel_blur_size:any=(55,5)
            blur = cv2.GaussianBlur(gray,kernel_blur_size,0)
            # invert the image 
            thresh=cv2.threshold(blur,150,250,cv2.THRESH_BINARY_INV+cv2.THRESH_OTSU)[1]
            # Dilate
            kernel_dilated_size:any=(55,10)
            kernel = cv2.getStructuringElement(cv2.MORPH_RECT,kernel_dilated_size)
            dilate = cv2.dilate(thresh, kernel, iterations=1)
            # find header and footer
            cnts = cv2.findContours(dilate,cv2.RETR_LIST,cv2.CHAIN_APPROX_TC89_KCOS)
            cnts = cnts[0] if len(cnts)==2 else cnts[1]
            for i,cnt in enumerate(cnts):
                x,y,w,h =cv2.boundingRect(cnt)
                if (y<=200 or y>=2000 )and w > 300 :
                    print(i,x,y,w,h)
                    print("footer/Header trouvé")
                    cv2.rectangle(image,(x,y),(x+w,y+h),(255,255,255),-1)
            final_image_path =f"./data/temp/proceeded_page{i}.png"
            cv2.imwrite(final_image_path,image)
            imgs.append(final_image_path)
            os.remove(img_path)
    # Save all converted images in a pdf
    # opening from filename
    # specify paper size (A4)
    a4inpt = (img2pdf.mm_to_pt(210),img2pdf.mm_to_pt(297))
    layout_A4 = img2pdf.get_layout_fun(a4inpt)
    with open(target,"wb+") as f:
        f.write(img2pdf.convert(imgs,layout_fun=layout_A4))
    # remove all tempory file
    empty_folder("./data/temp/*.png")
    if os.path.exists(target) is True :
        return 0, target
    return -3, f"Le fichier {target} n'a pu être créé"            
            



import logging

import tqdm
from lxf.settings import get_logging_level, nlp_with_vectors, enable_tqdm

#logger
logger = logging.getLogger('Tables Helper')
fh = logging.FileHandler('./logs/tables_helper.log')
fh.setLevel(get_logging_level())
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.setLevel(get_logging_level())
logger.addHandler(fh)

from lxf.domain.tables import lxfTables, lxfTable
from lxf.utilities.text_comparator import TextArrayComparator

def reconciliation_tables(tables:list[lxfTable])-> tuple[int, str,list[lxfTable]|None]:
    """
    Reconcilent plusieurs tables consécutives qui ont le même entête (cad 1 ligne de la table)
    """
    
    # Vérifications
    if tables==None or len (tables)==0 :
        return -100 , "Tables ne peut pas vide", None
    entete_courante = [cell.value for cell in tables[0].rows[0].cells]
    entete_comparator:TextArrayComparator=TextArrayComparator(entete_courante,nlp_with_vectors)
    merged_tables:list[lxfTable]= []
    table_courrante:lxfTable=lxfTable()
    # initialisation avec la première ligne (entete) du premier tableau
    table_courrante.rows.append(tables[0].rows[0])
    for table in tqdm.tqdm(tables,desc="Lecture des tables",disable=not enable_tqdm) :
        entete_a_verifier = [cell.value for cell in table.rows[0].cells]
        if entete_comparator.compare_to(entete_a_verifier)<0.85 : 
            # ce n'est pas la même entête, c'est donc un nouveau tableau
            # ajouter le tableau 
            merged_tables.append(table_courrante)
            # nouvelle table courrante
            table_courrante = lxfTable
            table_courrante.rows.append(table.rows[0])
            # nouvelle entête
            entete_courante = entete_a_verifier
            # Nouvelle entête comparator
            entete_comparator=None 
            entete_comparator=TextArrayComparator(entete_courante,nlp_with_vectors)            
        # il faut ajouter toutes les lignes suivantes de 1 à len(table) -1
        for row in table.rows[1:] : 
            table_courrante.rows.append(row)   
    if len(table_courrante.rows)>0: 
        merged_tables.append(table_courrante)
    return 0 , "", merged_tables
    
    
            



import datetime
import os
import logging
import re
from lxf.settings import get_logging_level, nlp_with_vectors
###################################################################
logger = logging.getLogger('repertoire_extractor')
fh = logging.FileHandler('./logs/repertoire_extractor.log')
fh.setLevel(get_logging_level())
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.setLevel(get_logging_level())
logger.addHandler(fh)
#################################################################



from lxf.domain.predictions import Predictions
from lxf.ai.text_analysis.default_text_analysis import summarize_chunks
from lxf.domain.keyswordsandphrases import KeysWordsAndPhrases
from lxf.domain.extracted_data import HIERARCHIE_DOCUMENT, HIERARCHIE_REPERTOIRE, Chunk, ChunkMetadata, ExtractedData, ExtractedMetadata
from lxf.domain.tables import lxfRow, lxfTable
from lxf.ai.classification.classifier import extract_text_and_table_from_file
from lxf.utilities.tables_helper import reconciliation_tables 
from lxf.domain.repertoire import Nature, RepertoireNotaire
from lxf.utilities.text_comparator import TextComparator
from dateutil.parser import parse




def consolidate_tables(tables:list[lxfTable]) -> list[lxfTable] :
    """
    il peut y avoir des enregistrements de répertoire sur plusieurs lignes dans la colonne Répertoire
    ce n'est pas acceptable, il faut fusionner avec la ligne précédente les lignes qui n'ont pas de numéro de répertoire
    ni date de signature mais du texte pour répertoire    
    """
    consolided_tables:list[lxfTable]=[]
    for table in tables : 
        consolided_table:lxfTable=lxfTable()
        previous_row:lxfRow = table.rows[0]
        # Toujours ajouter l'entête cad la première ligne
        consolided_table.rows.append(previous_row)
        for row in table.rows[1:] : 
            num_rep=row.cells[0].value
            if num_rep==None : num_rep=""
            date_sign=row.cells[1].value
            if date_sign==None : date_sign=""
            text=row.cells[3].value
            if text==None : text=""                    
            if num_rep.strip()=="" and date_sign.strip()=="" and text.strip()!="":
                # Il faut fusioner le text avec la ligne précédente
                previous_row.cells[3].value+=text
            else :
                if num_rep!="" and date_sign!="" and text!="" :
                    # Ajouter cette ligne
                    consolided_table.rows.append(row)
                    previous_row=row
        # ajouter la table consolidée
        consolided_tables.append(consolided_table)  
    return consolided_tables  

def extract_repertoires_from_tables(tables:list[lxfTable])->list[RepertoireNotaire]:
    """
    Retrouver les champs RepertoireNotaire dans la table en fonction de l'entête de la table
    """
    # Définition des outils de comparaison pour retrouver le mapping des champs
    num_rep_comparator:TextComparator=TextComparator("N°Répert.",nlp_with_vectors)
    signature_compator:TextComparator=TextComparator("Signature",nlp_with_vectors)
    brevet_minute_comparator:TextComparator=TextComparator("B/M",nlp_with_vectors)
    repertoire_comparator:TextComparator=TextComparator("Répertoire",nlp_with_vectors)
    support_comparator:TextComparator=TextComparator("Support",nlp_with_vectors)
    date_formalite_comparator:TextComparator=TextComparator("Date de formalité",nlp_with_vectors)
    depot_comparator:TextComparator=TextComparator("Dépôt", nlp_with_vectors)
    percu_comparator:TextComparator=TextComparator('Perçu',nlp_with_vectors)
    dt_etat_comparator:TextComparator=TextComparator("Dt sur Etat")
    visa:TextComparator=TextComparator("Visa",nlp_with_vectors)
    num_clerc_comparator:TextComparator=TextComparator("Numéro Clerc")
    # Clés pour le mapping des champs
    NUM_REP="num_rep"
    SIGNATURE="signature"
    BREVET_MINUTE="brevet_minute"
    TEXT="text"
    SUPPORT="support"
    DATE_FORMALITE="date_form"
    DEPOT="depot"
    PERCU="percu"
    DT_ETAT="dt_etat"
    VISA="visa"
    NUM_CLERC="num_clerc"   
     
    repertoires:list[RepertoireNotaire]=[]
    for table in tables:
        # Récupérer l'entête
        entete:list[str] = [cell.value for cell in table.rows[0].cells]
        # Retrouver le mapping des champs par rapport à l'entête
        mapping_fields:dict=dict()        
        for i,key in enumerate(entete) : 
            if num_rep_comparator.compare_to(key) >0.85 :
                mapping_fields[NUM_REP]=i
            elif signature_compator.compare_to(key)>0.85 :
                mapping_fields[SIGNATURE]=i
            elif brevet_minute_comparator.compare_to(key)>0.85 :
                mapping_fields[BREVET_MINUTE]=i                    
            elif repertoire_comparator.compare_to(key)>0.85 :
                mapping_fields[TEXT]=i                        
            elif support_comparator.compare_to(key)>0.85 :
                mapping_fields[SUPPORT]=i 
            elif date_formalite_comparator.compare_to(key)>0.85 :
                mapping_fields[DATE_FORMALITE]=i 
            elif depot_comparator.compare_to(key)>0.85 :
                mapping_fields[DEPOT]=i     
            elif percu_comparator.compare_to(key)>0.85 :
                mapping_fields[PERCU]=i       
            elif dt_etat_comparator.compare_to(key)>0.85 :
                mapping_fields[DT_ETAT]=i    
            elif visa.compare_to(key)>0.85 :
                mapping_fields[VISA]=i          
            elif num_clerc_comparator.compare_to(key)>0.85 :
                mapping_fields[NUM_CLERC]=i                                                                                                             
        # Mapping retrouvé
        # Mapper les champs en fonction de l'entête
        for row in table.rows[1:] : 
            # Le numéro de répertoire est obligatoire
            if row.cells[mapping_fields[NUM_REP]].value.strip()=="" : 
                logger.warning("Un enregistrement n'a pas de numéro de répertoire, il a été ignoré")
            else:                   
                rep:RepertoireNotaire=RepertoireNotaire()
                rep.num_repertoire = row.cells[mapping_fields[NUM_REP]].value
                rep.date_signature=row.cells[mapping_fields[SIGNATURE]].value
                b_m=row.cells[mapping_fields[BREVET_MINUTE]].value
                if b_m.lower()=="m" : rep.nature=Nature.Minute
                else : rep.nature = Nature.Brevet
                rep.text = row.cells[mapping_fields[TEXT]].value
                rep.support=row.cells[mapping_fields[SUPPORT]].value
                rep.date_formalite=row.cells[mapping_fields[DATE_FORMALITE]].value
                rep.depot=row.cells[mapping_fields[DEPOT]].value
                rep.droit_sur_etat=row.cells[mapping_fields[DT_ETAT]].value
                rep.visa=row.cells[mapping_fields[VISA]].value
                num_clerc:str= row.cells[mapping_fields[NUM_CLERC]].value
                tmp= num_clerc.split(" ")
                rep.numero_acte =tmp[0]
                rep.clerc=tmp[1]
                repertoires.append(rep)
    return repertoires

def get_extractedData_from_repertoires(repertoires:list[RepertoireNotaire],documentId:str, source:str,predictions:Predictions=Predictions()) -> ExtractedData :
    """
    Converti la liste des répertoires en ExtractedData
    """
    extracteData:ExtractedData=ExtractedData()
    extracteData.metadata = ExtractedMetadata()
    extracteData.metadata.origine="GeDAIA"
    extracteData.metadata.documentId=documentId
    extracteData.metadata.traite_le = datetime.datetime.today().strftime('%d/%m/%Y %H:%M')
    extracteData.keywords=["Répertoire","Minute","Brevet","Date Signature", "Date Formalité","Numéro répertoire", "Numéro d'acte","Support AAE","Support papier"]
    # Données de classification
    extracteData.metadata.classe=predictions.BestPrediction
    extracteData.metadata.classe_confidence=predictions.BestPredictionConfidence
    extracteData.explain="Ces données ont été extraites du répertoire de l'office depuis des documents PDF"      
    
    extracteData.chunks=[]
    for repertoire in repertoires : 
        chunk:Chunk=Chunk()
        chunk.metadata=ChunkMetadata()
        chunk.metadata.chunk=1
        chunk.metadata.chunks=1
        chunk.metadata.title = repertoire.text.split(":")[0][:80]
        chunk.metadata.source=source
        chunk.metadata.extra_metadata["Classification"]=predictions.BestPrediction
        chunk.metadata.extra_metadata["Classification confiance"]=predictions.BestPredictionConfidence
        try:
            if repertoire.date_signature!=None and repertoire.date_signature!="":
                 chunk.metadata.extra_metadata["date_signature"]=datetime.datetime.strptime(repertoire.date_signature,"%d/%m/%Y")
            if repertoire.date_formalite!=None and repertoire.date_formalite.strip()!="":
                chunk.metadata.extra_metadata["date_formalite"]=datetime.datetime.strptime(repertoire.date_formalite,"%d/%m/%Y")
        except Exception as ex:
            logger.critical(f"Tentative de conversion en date\nSignature={repertoire.date_signature}\nFormalite={repertoire.date_formalite}\nException: {ex}",stack_info=True)
        chunk.metadata.extra_metadata["num_rep"]=repertoire.num_repertoire    
        chunk.metadata.extra_metadata["nature"]=repertoire.nature
        chunk.metadata.extra_metadata["support"]=repertoire.support
        chunk.metadata.extra_metadata["num_acte"]=repertoire.numero_acte        
        chunk.metadata.extra_metadata["initiales_clerc"]=repertoire.clerc
        # Il s'agit du répertoire
        chunk.metadata.hierarchie=HIERARCHIE_REPERTOIRE
        
        
        chunk.page_content =f"""
Numéro de répertoire : {repertoire.num_repertoire}
Date de Signature : {repertoire.date_signature}
Nature : {repertoire.nature}
Texte:
{repertoire.text}
Support : {repertoire.support}
Date Formalité : {repertoire.date_formalite}
Dépôt : {repertoire.depot}
Perçu : {repertoire.frais_percus}
Droit sur Etat: {repertoire.droit_sur_etat}
Visa : {repertoire.visa}
Numéro acte dans la rédaction : {repertoire.numero_acte}
Initiales Clerc : {repertoire.clerc}        
        """ 
        
        # if len(repertoire.text)> 30 :
        #     tmp = summarize_chunks(repertoire.text[:1024]) 
        #     chunk_summaries = "".join(tmp)
        # else : 
        #     chunk_summaries=repertoire.text
        chunk.summary=repertoire.text[:256]        
        #chunk.summary=repertoire.text.split(":")[0]
        chunk.explain="Ces données ont été extraites du répertoire de l'office depuis des documents PDF"        
        kw_process=KeysWordsAndPhrases(chunk.page_content)
        if kw_process :
            keywords=kw_process.get_key_words(threshold=0.3)
            chunk.keywords=[key for key in keywords]                
        extracteData.chunks.append(chunk)
    return extracteData    
        
    
async def extract_repertoire_data(file_path:str,**kwargs)->tuple[int,str,ExtractedData|None]:
    """
    Extraction des données du répertoire contenu dans le fichier 
    """
    if file_path==None or file_path=="": 
        err_msg:str= "Argument vide ou None!"
        logging.error(err_msg,stack_info=True)
        return -2,err_msg, None
    if os.path.isfile(file_path) ==False :
        err_msg:str=f"Le fichier {file_path} n'existe pas!"
        logging.error(err_msg, stack_info=True)
        return -3, err_msg,None
    documentId:str=kwargs.get("documentId","")
    predictions:Predictions = kwargs.get("predictions",Predictions())
    try: 
        tables:list[lxfTable] 
        regex = r"(R[É|E]PERTOIRE *OFFICIEL)"
        # Rechercher d'abord dans le texte pour vérifier qu'il s'agit d'un répertoire officiel
        txt,_ = await extract_text_and_table_from_file(file_name=file_path,layout=False,extract_tables=False, extract_text=True)
        matches = re.findall(regex,txt[:1024],re.IGNORECASE)
        if matches == None  or len(matches)==0:
            return -4, "Ce n'est pas un répertoire officiel", None
        _, tables = await extract_text_and_table_from_file(file_name=file_path,layout=True,extract_tables=True, extract_text=False)
        if tables!=None :     
            err, err_msg, merged_tables = reconciliation_tables(tables)
            if err<0 :
                #Erreur 
                err_msg:str=f"Reconciliation des tableaux erreur : {err} - {err_msg}"
                logging.error(err_msg)
                return -4, err_msg, None                  
            # il peut y avoir des enregistrements de répertoire sur plusieurs lignes dans la colonne Répertoire
            # ce n'est pas acceptable, il faut fusionner avec la ligne précédente les lignes qui n'ont pas de numéro de répertoire
            # ni date de signature mais du texte pour répertoire
            consolided_tables = consolidate_tables(merged_tables)    
            repertoires:list[RepertoireNotaire] = extract_repertoires_from_tables(consolided_tables)
            if len(repertoires)==0 :
                return -4 , "Aucun répertoire extrait", None
            data:ExtractedData = get_extractedData_from_repertoires(repertoires,documentId, file_path, predictions)
            if data!=None :                 
                return 0, "", data
            else :
                logger.error(f"La conversion en ExtractedData a échouée",stack_info=True)
                return -10, "La conversion en ExtractedData a échouée"
    except Exception as excep:
        err_msg:str = f"Une erreur critique est apparue:\n{excep}"
        logging.critical(err_msg,stack_info=True)
        return -1, err_msg, None


from langchain_ollama import ChatOllama
from langchain_openai import ChatOpenAI
import re


class Interviewer():
    """
    Docstring for Interviewer
    """
    def __init__(self,
                 model:str="gpt-oss", 
                 base_url:str="http://localhos/11434",
                 ctx:int=4096,
                 system_prompt:str="Tu es un assistant français",
                 openapi_compatibility:bool=False, api_key:str="123" ):
        """
        Docstring for __init__
        
        :param self: Description
        :param model: Model LLM 
        :type model: str
        :param base_url: URL du serveur OLLAMA
        :type base_url: str
        :param ctx: Taille du contexte (défaut 4096 o)
        :type ctx: int
        """
        if openapi_compatibility : 
            self.llm = ChatOpenAI(
                model=model,
                temperature=0,
                max_completion_tokens=ctx,
                base_url=base_url,
                api_key=api_key,
            )
        else :
            
            self.llm = ChatOllama(
                model=model,
                temperature=0,
                num_ctx=ctx,
                base_url=base_url,
                validate_model_on_init=True,
                
                # other params...
            )
        self.system_prompt=system_prompt
        
    def invoke(self,user_prompt:str)->str :
        """
        """
        messages = [
            ("system",self.system_prompt),
            ("human",user_prompt)
        ]
        ai_response =self.llm.invoke(messages)
        if ai_response : 
            return ai_response.content
        return "Aucune réponse"

    async def invoke_async(self,user_prompt:str)->tuple[int,str] :
        """
        """
        messages = [
            ("system",self.system_prompt),
            ("human",user_prompt)
        ]
        ai_response =await self.llm.ainvoke(messages)
        if ai_response : 
            return 0,ai_response.content
        return -1,"Aucune réponse"



import logging
from typing import List
from lxf.settings import get_logging_level
logger = logging.getLogger('Repertoire Office')
fh = logging.FileHandler('./logs/repertoire_office.log')
fh.setLevel(get_logging_level())
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.setLevel(get_logging_level())
logger.addHandler(fh)
from pydantic import BaseModel,Field
from enum import StrEnum

class Nature(StrEnum):
    Brevet="BREVET"
    Minute="MINUTE"
    
class RepertoireNotaire(BaseModel):
    """
    Enregistrement d'un acte au répertoire
    """
    num_repertoire:str=Field(default="")
    date_signature:str=Field(default="")
    nature:Nature=Field(default=Nature.Minute)
    text:str=Field(default="")
    support:str=Field(default="")
    date_formalite:str=Field(default="")
    depot:str=Field(default="")
    frais_percus:str=Field(default="")
    droit_sur_etat:str=Field(default="")
    visa:str=Field(default="")
    numero_acte:str=Field(default="")
    clerc:str=Field(default="")